<?php

namespace Taxus;

class Taxus
{
    # @ArrayOf Privilege
    protected array $list = [];

    # @ArrayOf PermissionGrantInterface
    protected array $gates = [];

    public function __construct(PermissionGrantInterface ... $gate,)
    {
        $this->pushGate(... $gate);
    }

    public function pushGate(PermissionGrantInterface ... $gates) : void
    {
        $this->gates = array_merge($this->gates, $gates);
    }

    public function hasPrivilege(string $name) : bool
    {
        return isset($this->list[$name]);
    }

    public function add(...$privileges) : self
    {
        foreach($privileges as $item) {
            $item = (array) $item;

            if ( ! $item[0] instanceof Privilege ) {
                throw new \InvalidArgumentException("First element must be an instance of '" . Privilege::class . "'");
            }

            if ( ( $item[1] ?? false ) && ! is_string($item[1]) ) {
                throw new \InvalidArgumentException("Second element must be a callable function from your Permission Handler");
            }

            $this->list[ $item[0]->name ] = $item;
        }

        return $this;
    }

    /**
     * @param $name
     * @param ...$arguments
     * @return bool|null Return TRUE if granted, FALSE if not and NULL if this call should be skipped
     */
    public function granted(\BackedEnum|string|null $name, ...$arguments) : bool|TaxusGrantEnum
    {
        if ($name === null) {
            return true;
        }

        $name = $name instanceof \BackedEnum ? $name->value : $name;

        if ( ! isset($this->list[$name]) ) {
            throw new \InvalidArgumentException("Privilege '{$name}' could not be found. Did you forgot to add it to your Taxus object ?");
        }

        $obj = $this->list[$name][0];

        $callback = $this->list[$name][1] ?? 'default';

        foreach($this->gates as $gate) {
            if ( method_exists($gate, $callback) ) {
                return $gate->$callback(...$arguments);
            }
        }

        throw new \Exception(
            sprintf(
                "Method '$callback' could not be found from given gates list: %s",
                implode(', ', array_map(fn($e) => $e::class, $this->gates))
            )
        );
    }
}
